package org.broadinstitute.sting.gatk.walkers.SMUGwalkers;

import net.sf.samtools.SAMReadGroupRecord;
import org.broadinstitute.sting.commandline.Argument;
import org.broadinstitute.sting.commandline.ArgumentCollection;
import org.broadinstitute.sting.commandline.Output;
import org.broadinstitute.sting.commandline.RodBinding;
import org.broadinstitute.sting.gatk.DownsampleType;
import org.broadinstitute.sting.gatk.arguments.DbsnpArgumentCollection;
import org.broadinstitute.sting.gatk.contexts.AlignmentContext;
import org.broadinstitute.sting.gatk.contexts.AlignmentContextUtils;
import org.broadinstitute.sting.gatk.contexts.ReferenceContext;
import org.broadinstitute.sting.gatk.filters.BadMateFilter;
import org.broadinstitute.sting.gatk.filters.MappingQualityUnavailableFilter;
import org.broadinstitute.sting.gatk.refdata.RefMetaDataTracker;
import org.broadinstitute.sting.gatk.walkers.*;
import org.broadinstitute.sting.gatk.walkers.annotator.VariantAnnotatorEngine;
import org.broadinstitute.sting.gatk.walkers.annotator.interfaces.AnnotatorCompatibleWalker;
import org.broadinstitute.sting.gatk.walkers.genotyper.UnifiedArgumentCollection;
import org.broadinstitute.sting.gatk.walkers.genotyper.UnifiedGenotyperEngine;
import org.broadinstitute.sting.gatk.walkers.genotyper.VariantCallContext;
import org.broadinstitute.sting.utils.BaseUtils;
import org.broadinstitute.sting.utils.GenomeLoc;
import org.broadinstitute.sting.utils.pileup.ReadBackedPileup;
import org.broadinstitute.sting.utils.variantcontext.Genotype;
import org.broadinstitute.sting.utils.variantcontext.VariantContext;
import org.broadinstitute.sting.utils.variantcontext.VariantContextUtils;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.*;

/**
 * Created by Zhuo Song and Chun Li.
 * User: zsong
 * Date: 5/22/12
 * Time: 12:12 PM
 * To change this template use File | Settings | File Templates.
 */
@ReadFilters( {BadMateFilter.class, MappingQualityUnavailableFilter.class} )
@Reference(window=@Window(start=-200,stop=200))
@By(DataSource.REFERENCE)
@Downsample(by= DownsampleType.BY_SAMPLE, toCoverage=250)
public class SMUG extends LocusWalker<Integer[], Integer[]> implements AnnotatorCompatibleWalker {

    @Output
    PrintStream out;

    @ArgumentCollection
    private UnifiedArgumentCollection UAC = new UnifiedArgumentCollection();
    @ArgumentCollection
    protected DbsnpArgumentCollection dbsnp = new DbsnpArgumentCollection();
    public RodBinding<VariantContext> getDbsnpRodBinding() { return dbsnp.dbsnp; }
    public RodBinding<VariantContext> getSnpEffRodBinding() { return null; }
    public List<RodBinding<VariantContext>> getCompRodBindings() { return Collections.emptyList(); }
    public List<RodBinding<VariantContext>> getResourceRodBindings() { return Collections.emptyList(); }
    public boolean alwaysAppendDbsnpId() { return false; }

    @Argument(fullName = "debug_file", shortName = "debug_file", doc = "File to print all of the annotated and detailed debugging output", required = false)
    protected PrintStream verboseWriter = null;
    @Argument(fullName="annotation", shortName="A", doc="One or more specific annotations to apply to variant calls", required=false)
    protected List<String> annotationsToUse = new ArrayList<String>();
    @Argument(fullName="excludeAnnotation", shortName="XA", doc="One or more specific annotations to exclude", required=false)
    protected List<String> annotationsToExclude = new ArrayList<String>();
    @Argument(fullName="group", shortName="G", doc="One or more classes/groups of annotations to apply to variant calls", required=false)
    protected String[] annotationClassesToUse = { "Standard" };
    @Argument(fullName = "SampleName_Tumor", shortName = "smT", doc = "sample name of tumor tissue", required = true)
    String smT;
    @Argument(fullName = "SampleName_Normal", shortName = "smN", doc = "sample name of normal tissue", required = true)
    String smN;
    @Argument(fullName = "SampleType_Tumor", shortName = "typeT", doc = "sample type of tumor tissue", required = false)
    String typeT       = "Tumor";
    @Argument(fullName = "SampleType_Normal", shortName = "typeN", doc = "sample type of normal tissue", required = false)
    String typeN       = "Normal";
    @Argument(fullName = "minBaseQ", shortName = "minBQ", doc = "min base quality, default=20", required = false)
    int minBQ          = 20;
    @Argument(fullName = "minMappingQ",  shortName = "minMQ", doc = "min mapping quality, default=20", required = false)
    int minMQ          = 20;
    @Argument(fullName = "minAllAlleles_Normal", shortName = "minAllN", doc = "min sum of all base counts in Normal after completely filtering (BQ,MQ) of the sample, default=10", required = false)
    int minAllNormal   = 10;
    @Argument(fullName = "minAllAlleles_Tumor", shortName = "minAllT", doc = "min sum of all base counts in Tumor after completely filtering (BQ,MQ) of the sample, default=5", required = false)
    int minAllTumor    = 10;
    @Argument(fullName = "minMinorAlleles_Tumor", shortName = "minRestT", doc = "min sum of three minor alleles' base counts in Tumor after completely filtering (BQ,MQ) of the sample, default=2", required = false)
    int minRestTumor   = 2;
    @Argument(fullName = "minHeteroRate_Tumor", shortName = "minHRT", doc = "min hetero rate in Tumor: Minors/All, default=0.1", required = false)
    double minHRT      = 0.01;
    //@Argument(fullName = "minLOHScore_Tumor", shortName = "minLOHScore", doc = "min LOH score in Tumor: phredBinomial - E(phred), default=30", required = false)
    //double minLOHScore = 30;
    /*   GATK:
         Since we cap each base quality (BQ) by its read's mapping quality (MQ) [in Read.updateBaseAndQuality()], then:
         if minBQ > minMQ, then we require that MQ be >= minBQ as well.
         [Otherwise, we end up capping BQ by MQ only AFTER we tried removing bases with BQ < minBQ, which is WRONG!]
         To do this properly, we set: minMQ = max(minMQ, minBQ)
    */
    protected PrintStream BaseSub = null;
    protected PrintStream LOH     = null;
    private Set<String> _sampleNames = new HashSet<String>();
    private Set<String> _sampleName1 = new HashSet<String>();
    private Set<String> _sampleName2 = new HashSet<String>();
    private int _isTExist = 0;
    private int _isBExist = 0;
    private String _BSf = null;
    private String _LOHf = null;
    /*  GATK:
        //the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls. This argument is passed to the Unified Genotyper.
        @Argument(fullName="standard_min_confidence_threshold_for_calling", shortName="stand_call_conf", doc="the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls", required=false)
        private double callConf = -1;
        //@Argument(fullName ="standard_min_confidence_threshold_for_calling", shortName="callConf", doc="the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls, default=30", required=false)
        //double callConf    = 30;
    */
    private UnifiedGenotyperEngine UG_engine1 = null;
    private UnifiedGenotyperEngine UG_engine2 = null;
    private VariantAnnotatorEngine annotationEngine;
    double call_Conf = 0.0;
    int    call_mbq  = 0;

    public void initialize() {
        call_Conf = UAC.STANDARD_CONFIDENCE_FOR_CALLING;
        call_mbq  = UAC.MIN_BASE_QUALTY_SCORE;
        List<SAMReadGroupRecord> read_groups = this.getToolkit().getSAMFileHeader().getReadGroups();
        for ( SAMReadGroupRecord record : read_groups ) {
            String sample = record.getSample();
            if ( sample != null )
                _sampleNames.add(sample);
        }
        for ( String sample : _sampleNames) {
            if (sample.equals(smT)) {
                _isTExist = 1;
                _sampleName1.add(smT);
            }
            if (sample.equals(smN)) {
                _isBExist = 1;
                _sampleName2.add(smN);
            }
            //out.println("## " + sample);
        }
        String _error1 = "## ERROR 1:\t" + smT + " doesn't exist in data";
        if (_isTExist == 0) {
            System.err.println(_error1);
            System.exit(0);
        }
        String _error2 = "## ERROR 2:\t" + smN + " doesn't exist in data";
        if (_isBExist == 0) {
            System.err.println(_error2);
            System.exit(0);
        }
        //
        if ( verboseWriter != null )
            verboseWriter.println("AFINFO\tLOC\tREF\tALT\tMAF\tF\tAFprior\tAFposterior\tNormalizedPosterior");
        annotationEngine = new VariantAnnotatorEngine(Arrays.asList(annotationClassesToUse), annotationsToUse, annotationsToExclude, this, getToolkit());
        UG_engine1 = new UnifiedGenotyperEngine(getToolkit(), UAC, logger, verboseWriter, annotationEngine, _sampleName1, VariantContextUtils.DEFAULT_PLOIDY);
        UG_engine2 = new UnifiedGenotyperEngine(getToolkit(), UAC, logger, verboseWriter, annotationEngine, _sampleName2, VariantContextUtils.DEFAULT_PLOIDY);
        //
        _BSf  = smN + "." + typeN + "." + smT + "." + typeT + "." + "ToAnalyzeBaseSub.txt";
        _LOHf = smN + "." + typeN + "." + smT + "." + typeT + "." + "ToAnalyzeLOH.txt";
        //
        FileOutputStream _BS = null;
        try {
            _BS = new FileOutputStream(_BSf);
        } catch (FileNotFoundException e) {
            e.printStackTrace(System.err);
        }
        BaseSub = new PrintStream(_BS);
        String BaseSubHeader = "#loc\trsID\tref\tgt_N\tgt_T\tgtQ_N\tgtQ_T\tN_hR\tT_hR\tN_hR_C\tT_hR_C\tN_aC\tT_aC\tN_+-\tT_+-\tN_filtered\tT_filtered";
        BaseSub.println(BaseSubHeader);
        //
        FileOutputStream _LOH = null;
        try {
            _LOH = new FileOutputStream(_LOHf);
        } catch (FileNotFoundException e) {
            e.printStackTrace(System.err);
        }
        LOH = new PrintStream(_LOH);
        String LOHHeader = "#loc\trsID\tref\tgt_N\tgt_T\tgtQ_N\tgtQ_T\tLOH_Score\tN_aC\tT_aC\tT_Major_aC\tT_Minor_aC\tN_+-\tT_+-\tN_filtered\tT_filtered";
        LOH.println(LOHHeader);

    }

    @Override
    public Integer[] map(RefMetaDataTracker tracker, ReferenceContext ref, AlignmentContext context) {
        int _nPotentialBaseSub = 0;
        int _nPotentialLOH     = 0;
        DecimalFormat df4 = new DecimalFormat("#.####");
        DecimalFormat df2 = new DecimalFormat("#.##");
        //
        Map<String, AlignmentContext> _smT_ContextMap;
        Map<String, AlignmentContext> _smN_ContextMap;
        //
        if ( BaseUtils.simpleBaseToBaseIndex(ref.getBase()) != -1 ) {
            if ( ! context.hasExtendedEventPileup() ) {
                _smT_ContextMap = AlignmentContextUtils.splitContextBySampleName(context, smT);
                _smN_ContextMap = AlignmentContextUtils.splitContextBySampleName(context, smN);
                if (_smT_ContextMap != null && _smN_ContextMap != null) {
                    // before filter BQ, MQ; GenotypeCall based on this
                    AlignmentContext _smT_Context = _smT_ContextMap.get(smT);
                    AlignmentContext _smN_Context = _smN_ContextMap.get(smN);
                    //
                    if (_smT_Context != null && _smN_Context != null) {
                        ReadBackedPileup _smT_Context_COMPLETE_RBP = _smT_Context.getBasePileup();
                        ReadBackedPileup _smN_Context_COMPLETE_RBP = _smN_Context.getBasePileup();
                        ReadBackedPileup _smT_Context_FORWARD_RBP  = _smT_Context.getBasePileup().getPositiveStrandPileup();
                        ReadBackedPileup _smN_Context_FORWARD_RBP  = _smN_Context.getBasePileup().getPositiveStrandPileup();
                        ReadBackedPileup _smT_Context_REVERSE_RBP  = _smT_Context.getBasePileup().getNegativeStrandPileup();
                        ReadBackedPileup _smN_Context_REVERSE_RBP  = _smN_Context.getBasePileup().getNegativeStrandPileup();
                        int _nsmTcc = _smT_Context_COMPLETE_RBP.depthOfCoverage();
                        int _nsmNcc = _smN_Context_COMPLETE_RBP.depthOfCoverage();
                        GenomeLoc _location = _smT_Context_COMPLETE_RBP.getLocation();
                        //
                        if (_nsmTcc >= 1 && _nsmNcc >= 1) {
                            ReadBackedPileup _smT_COMPLETE_Filtered_BQMQ = _smT_Context_COMPLETE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _smN_COMPLETE_Filtered_BQMQ = _smN_Context_COMPLETE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_CompleteF                          = _smT_COMPLETE_Filtered_BQMQ.depthOfCoverage();
                            int _nsmN_CompleteF                          = _smN_COMPLETE_Filtered_BQMQ.depthOfCoverage();
                            int[] _smT_CompleteFBC                       = _smT_COMPLETE_Filtered_BQMQ.getBaseCounts();
                            int[] _smN_CompleteFBC                       = _smN_COMPLETE_Filtered_BQMQ.getBaseCounts();
                            int _aT = _smT_CompleteFBC[0], _cT = _smT_CompleteFBC[1], _gT = _smT_CompleteFBC[2], _tT = _smT_CompleteFBC[3];
                            int _aN = _smN_CompleteFBC[0], _cN = _smN_CompleteFBC[1], _gN = _smN_CompleteFBC[2], _tN = _smN_CompleteFBC[3];
                            ReadBackedPileup _SM1_FORWARD_FilteredBQMQ = _smT_Context_FORWARD_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _SM2_FORWARD_FilteredBQMQ = _smN_Context_FORWARD_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_ForwardF                         = _SM1_FORWARD_FilteredBQMQ.depthOfCoverage();
                            int _nsmN_ForwardF                         = _SM2_FORWARD_FilteredBQMQ.depthOfCoverage();
                            int[] _nsmT_ForwardFBC                     = _SM1_FORWARD_FilteredBQMQ.getBaseCounts();
                            int[] _nsmN_ForwardFBC                     = _SM2_FORWARD_FilteredBQMQ.getBaseCounts();
                            int _afT = _nsmT_ForwardFBC[0], _cfT = _nsmT_ForwardFBC[1], _gfT = _nsmT_ForwardFBC[2], _tfT = _nsmT_ForwardFBC[3];
                            int _afN = _nsmN_ForwardFBC[0], _cfN = _nsmN_ForwardFBC[1], _gfN = _nsmN_ForwardFBC[2], _tfN = _nsmN_ForwardFBC[3];
                            ReadBackedPileup _smT_REVERSE_FilteredBQMQ = _smT_Context_REVERSE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _smN_REVERSE_FilteredBQMQ = _smN_Context_REVERSE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_ReverseF                         = _smT_REVERSE_FilteredBQMQ.depthOfCoverage();
                            int _nsmN_ReverseF                         = _smN_REVERSE_FilteredBQMQ.depthOfCoverage();
                            int[] _nsmT_ReverseFBC                     = _smT_REVERSE_FilteredBQMQ.getBaseCounts();
                            int[] _nsmN_ReverseFBC                     = _smN_REVERSE_FilteredBQMQ.getBaseCounts();
                            int _arT = _nsmT_ReverseFBC[0], _crT = _nsmT_ReverseFBC[1], _grT = _nsmT_ReverseFBC[2], _trT = _nsmT_ReverseFBC[3];
                            int _arN = _nsmN_ReverseFBC[0], _crN = _nsmN_ReverseFBC[1], _grN = _nsmN_ReverseFBC[2], _trN = _nsmN_ReverseFBC[3];
                            int[] _T  = {_aT,_cT,_gT,_tT};
                            int[] _N  = {_aN,_cN,_gN,_tN};
                            int _All_T = 0;
                            int _All_N = 0;
                            int i;
                            for (i = 0; i < 4; i++) {
                                _All_T += _T[i];
                                _All_N += _N[i];
                            }
                            //
                            AlignmentContext _smT_Context_BQMQ = new AlignmentContext(_location, _smT_COMPLETE_Filtered_BQMQ);
                            AlignmentContext _smN_Context_BQMQ = new AlignmentContext(_location, _smN_COMPLETE_Filtered_BQMQ);
                            //VariantCallContext _vcT = UG_engine1.calculateLikelihoodsAndGenotypes(tracker, ref, _smT_Context);
                            //VariantCallContext _vcN = UG_engine2.calculateLikelihoodsAndGenotypes(tracker, ref, _smN_Context);
                            VariantCallContext _vcT = UG_engine1.calculateLikelihoodsAndGenotypes(tracker, ref, _smT_Context_BQMQ).get(0);
                            VariantCallContext _vcN = UG_engine2.calculateLikelihoodsAndGenotypes(tracker, ref, _smN_Context_BQMQ).get(0);
                            //
                            String _refAllele = new String(new byte[]{ref.getBase()});
                            String _rsID = ".";
                            if (_vcT.hasID()) {
                                _rsID = _vcT.getID();
                            }
                            /********************/
                            //String _s0 = _location+"\t"+_rsID+"\t"+_refAllele+"\t"+_aN+","+_cN+","+_gN+","+_tN+"\t"+_aT+","+_cT+","+_gT+","+_tT+"\t"+_vcN.toString()+"\t"+_vcT.toString();
                            //out.println(_s0);
                            /********************/
                            Genotype _gt_N = null;
                            String  _gtS_N = "%";
                            double  _gtQ_N = -100.0;
                            if (_vcN.hasGenotype(smN)) {
                                 _gt_N =  _vcN.getGenotype(smN);
                                _gtS_N = _gt_N.getGenotypeString(false);
                                _gtQ_N = _gt_N.getPhredScaledQual();
                            }
                            Genotype _gt_T = null;
                            String  _gtS_T = "%";
                            double  _gtQ_T = -100.0;
                            if (_vcT.hasGenotype(smT)) {
                                 _gt_T =  _vcT.getGenotype(smT);
                                _gtS_T = _gt_T.getGenotypeString(false);
                                _gtQ_T = _gt_T.getPhredScaledQual();
                            }
                            //
                            String _lineGT = _location
                                    + "\t" + _rsID
                                    + "\t" + _refAllele
                                    + "\t" + _gtS_N
                                    + "\t" + _gtS_T
                                    + "\t" + df2.format(_gtQ_N)
                                    + "\t" + df2.format(_gtQ_T)
                                    ;
                            /********** BaseSub **********/
                            Map<String, Integer> map_T = new HashMap<String, Integer>();
                            Map<String, Integer> map_N = new HashMap<String, Integer>();
                            map_T.put("A",_aT);
                            map_T.put("C",_cT);
                            map_T.put("G",_gT);
                            map_T.put("T",_tT);
                            map_N.put("A",_aN);
                            map_N.put("C",_cN);
                            map_N.put("G",_gN);
                            map_N.put("T",_tN);
                            int BaseSub_Report   = 0;
                            boolean isNormalHomo = false;
                            boolean isTumorVar   = false;
                            String _allele_N     = "*";
                            String N_homo        = "N";
                            if (_All_T >= minAllTumor && _All_N >= minAllNormal) {
                                if (_vcN.getNAlleles() == 1) {
                                    isNormalHomo = true;
                                    // _allele_N = C*
                                    _allele_N = _vcN.getAlleles().get(0).toString();
                                } else if (_vcN.getNAlleles() > 1) {
                                    if (_vcN.hasGenotype(smN)) {
                                        if (_gt_N.isHom() && _gtQ_N >= call_Conf) {
                                            isNormalHomo = true;
                                            // _allele_N = G/G
                                            _allele_N = _gtS_N;
                                        }
                                    } else {
                                        out.println("## WEIRD: "+_location+","+_rsID+" (Normal) with >1 allels but doesn't have genotype ##");
                                    }
                                }
                            }
                            // genotype of Normal could be '%' NO_VARIATION or 'G/G' high qual homo, but not '.' NO_CALL
                            if (isNormalHomo) {
                                N_homo = _allele_N.substring(0,1);
                                //out.println(_allele_N+"("+N_homo+")");
                                int _varChrCount_N = 0;
                                int _allChrCount_N = 0;
                                for (String key : map_N.keySet()) {
                                    if (!key.equalsIgnoreCase(N_homo)) {
                                        _varChrCount_N += map_N.get(key);
                                    }
                                    _allChrCount_N += map_N.get(key);
                                }
                                int _varChrCount_T = 0;
                                int _allChrCount_T = 0;
                                for (String key : map_T.keySet()) {
                                    if (!key.equalsIgnoreCase(N_homo)) {
                                        _varChrCount_T += map_T.get(key);
                                    }
                                    _allChrCount_T += map_T.get(key);
                                }
                                //out.println(_varChrCount_N+"/"+_allChrCount_N+"\t"+_varChrCount_T+"/"+_allChrCount_T);
                                //
                                double _HeteroR_T = 0.0;
                                double _HeteroR_N = 0.0;
                                _HeteroR_T = (double)_varChrCount_T/(double)_allChrCount_T;
                                _HeteroR_N = (double)_varChrCount_N/(double)_allChrCount_N;
                                if (_varChrCount_T >= minRestTumor && _HeteroR_T >= minHRT) {
                                    isTumorVar = true;
                                }
                                if (isTumorVar) {
                                    BaseSub_Report = 1;
                                }
                                //
                                if (BaseSub_Report > 0) {
                                    String _lineBaseSub = df4.format(_HeteroR_N)
                                            + "\t" + df4.format(_HeteroR_T)
                                            + "\t" + _varChrCount_N + "/" + _allChrCount_N
                                            + "\t" + _varChrCount_T + "/" + _allChrCount_T
                                            + "\t" + _aN + "," + _cN + "," + _gN + "," + _tN
                                            + "\t" + _aT + "," + _cT + "," + _gT + "," + _tT
                                            + "\t" + _afN + "," + _arN + ";" + _cfN + "," + _crN + ";" + _gfN + "," + _grN + ";" + _tfN + "," + _trN
                                            + "\t" + _afT + "," + _arT + ";" + _cfT + "," + _crT + ";" + _gfT + "," + _grT + ";" + _tfT + "," + _trT
                                            + "\t" + smN + "(" + typeN + "):" + _nsmNcc + "->" + _nsmN_CompleteF + "=" + "(" + _nsmN_ForwardF + "+," + _nsmN_ReverseF + "-)"
                                            + "\t" + smT + "(" + typeT + "):" + _nsmTcc + "->" + _nsmT_CompleteF + "=" + "(" + _nsmT_ForwardF + "+," + _nsmT_ReverseF + "-)"
                                            ;
                                    String _outLine = _lineGT + "\t" + _lineBaseSub;
                                    BaseSub.println(_outLine);
                                    _nPotentialBaseSub++;
                                }
                            }
                            /********** LOH **********/
                            Arrays.sort(_T);
                            int _Major_T = _T[3];
                            int _Minor_T = _T[2];
                            double LOH_Score = 0.0;
                            int LOH_Report   = 0;
                            if (_All_T >= minAllTumor && _All_N >= minAllNormal) {
                                if (_vcN.hasGenotype(smN)) {
                                    if (_gt_N.isHet() && _gtQ_N >= call_Conf) {
                                        LOH_Score = getBinomialPhred(_Minor_T, _Major_T + _Minor_T, 0.5);
                                        //out.println("LOH_Score: "+df2.format(LOH_Score));
                                        //if (LOH_Score >= minLOHScore) {
                                            LOH_Report = 1;
                                        //}
                                    }
                                }
                            }
                            //
                            if (LOH_Report > 0) {
                                String _lineLOH = _aN + "," + _cN + "," + _gN + "," + _tN
                                        + "\t" + _aT + "," + _cT + "," + _gT + "," + _tT
                                        + "\t" + _Major_T + "\t" + _Minor_T
                                        + "\t" + _afN + "," + _arN + ";" + _cfN + "," + _crN + ";" + _gfN + "," + _grN + ";" + _tfN + "," + _trN
                                        + "\t" + _afT + "," + _arT + ";" + _cfT + "," + _crT + ";" + _gfT + "," + _grT + ";" + _tfT + "," + _trT
                                        + "\t" + smN + "(" + typeN + "):" + _nsmNcc + "->" + _nsmN_CompleteF + "=" + "(" + _nsmN_ForwardF + "+," + _nsmN_ReverseF + "-)"
                                        + "\t" + smT + "(" + typeT + "):" + _nsmTcc + "->" + _nsmT_CompleteF + "=" + "(" + _nsmT_ForwardF + "+," + _nsmT_ReverseF + "-)"
                                        ;
                                String _outLine = _lineGT + "\t" + df2.format(LOH_Score) + "\t" + _lineLOH;
                                LOH.println(_outLine);
                                _nPotentialLOH++;
                            }

                        }
                    }
                }
            }
        }
        Integer[] re = new Integer[2];
        re[0] = _nPotentialBaseSub;
        re[1] = _nPotentialLOH;
        return re;
    }

    @Override
    public Integer[] reduceInit() {
        Integer[] statis = new Integer[2];
        statis[0] = 0;
        statis[1] = 0;
        return statis;
    }

    @Override
    public Integer[] reduce(Integer[] map, Integer[] reduce) {
        reduce[0] += map[0];
        reduce[1] += map[1];
        return reduce;
    }

    public void onTraversalDone(Integer[] SMUGstatis) {
        out.println("## SMUG ## ---------------------------------------------------------------------");
        out.printf( "## SMUG ## typeN %s, typeT %s, minBQ %d, minMQ %d, minAllNormal %d, minAllTumor %d, minRestTumor %d, minHRT %.2f, callConf %.2f%n", typeN, typeT, minBQ, minMQ, minAllNormal, minAllTumor, minRestTumor, minHRT, call_Conf);
        out.println("## SMUG ## " + _BSf);
        out.println("## SMUG ## " + _LOHf);
        out.println("## SMUG ## The No. of loci with potential NormalHomo TumorHetero (BaseSub) is: " + SMUGstatis[0]);
        out.println("## SMUG ## The No. of loci with potential NormalHetero TumorHomo (LOH)     is: " + SMUGstatis[1]);
        out.println("## SMUG ## ---------------------------------------------------------------------");
    }

    private double getBinomialPhred(int k, int n, double p) {
        double [] ePhred99 = setEPhred99();
        int d = Math.min(n, 500);
        double pV = (n == 2 * k) ? 1 : 2 * pBinom(k, n, p) - dBinom(k, n, p);
        double pV_phred = -10 * Math.log10(pV);
        //out.println("pBinom: "+pBinom(k, n, p)+"\tdBinom: "+dBinom(k, n, p)+"\tpV: "+pV+"\tpV_phred: "+pV_phred+"\tePhred: "+ePhred99[d - 1]);
        long CAP = 99;
        return Math.min(pV_phred,CAP) - ePhred99[d - 1];
    }

    // x*(x-1)*..*y
    private double Factorial(int x, int y) {
        if (x == y - 1) {
            return 1;
        } else {
            return x * Factorial(x - 1, y);
        }
    }

    private double dBinom(int k, int n, double p) {
        if (k == 0) {
            return Math.pow((1 - p), n);
        } else {
            return (Factorial(n, n - k + 1) / Factorial(k,1)) * Math.pow(p, k) * Math.pow((1 - p), (n - k));
        }
    }

    private double pBinom(int k, int n, double p) {
        int i;
        double cdf = 0;
        for (i = k; i >= 0; i--) {
            //out.println(dBinom(i, n, p));
            cdf += dBinom(i, n, p);
        }
        return cdf;
    }

    private double[] setEPhred99() {
        double [] ePhred = new double[500];
        ePhred[0]	= 3.000;
        ePhred[1]	= 3.000;
        ePhred[2]	= 3.750;
        ePhred[3]	= 3.500;
        ePhred[4]	= 4.375;
        ePhred[5]	= 3.656;
        ePhred[6]	= 3.828;
        ePhred[7]	= 3.906;
        ePhred[8]	= 4.242;
        ePhred[9]	= 4.121;
        ePhred[10]	= 4.050;
        ePhred[11]	= 3.978;
        ePhred[12]	= 4.193;
        ePhred[13]	= 4.011;
        ePhred[14]	= 4.168;
        ePhred[15]	= 4.148;
        ePhred[16]	= 4.172;
        ePhred[17]	= 4.145;
        ePhred[18]	= 4.326;
        ePhred[19]	= 4.193;
        ePhred[20]	= 4.353;
        ePhred[21]	= 4.351;
        ePhred[22]	= 4.390;
        ePhred[23]	= 4.379;
        ePhred[24]	= 4.519;
        ePhred[25]	= 4.448;
        ePhred[26]	= 4.305;
        ePhred[27]	= 4.077;
        ePhred[28]	= 4.230;
        ePhred[29]	= 4.152;
        ePhred[30]	= 4.365;
        ePhred[31]	= 4.109;
        ePhred[32]	= 4.251;
        ePhred[33]	= 4.179;
        ePhred[34]	= 4.384;
        ePhred[35]	= 4.181;
        ePhred[36]	= 4.403;
        ePhred[37]	= 4.315;
        ePhred[38]	= 4.508;
        ePhred[39]	= 4.342;
        ePhred[40]	= 4.628;
        ePhred[41]	= 4.299;
        ePhred[42]	= 4.326;
        ePhred[43]	= 4.231;
        ePhred[44]	= 4.348;
        ePhred[45]	= 4.261;
        ePhred[46]	= 4.415;
        ePhred[47]	= 4.181;
        ePhred[48]	= 4.452;
        ePhred[49]	= 4.200;
        ePhred[50]	= 4.537;
        ePhred[51]	= 4.309;
        ePhred[52]	= 4.365;
        ePhred[53]	= 4.351;
        ePhred[54]	= 4.369;
        ePhred[55]	= 4.245;
        ePhred[56]	= 4.213;
        ePhred[57]	= 4.251;
        ePhred[58]	= 4.260;
        ePhred[59]	= 4.304;
        ePhred[60]	= 4.323;
        ePhred[61]	= 4.232;
        ePhred[62]	= 4.196;
        ePhred[63]	= 4.288;
        ePhred[64]	= 4.252;
        ePhred[65]	= 4.312;
        ePhred[66]	= 4.206;
        ePhred[67]	= 4.230;
        ePhred[68]	= 4.129;
        ePhred[69]	= 4.281;
        ePhred[70]	= 4.208;
        ePhred[71]	= 4.344;
        ePhred[72]	= 4.261;
        ePhred[73]	= 4.284;
        ePhred[74]	= 4.159;
        ePhred[75]	= 4.353;
        ePhred[76]	= 4.239;
        ePhred[77]	= 4.328;
        ePhred[78]	= 4.164;
        ePhred[79]	= 4.300;
        ePhred[80]	= 4.155;
        ePhred[81]	= 4.363;
        ePhred[82]	= 4.197;
        ePhred[83]	= 4.415;
        ePhred[84]	= 4.126;
        ePhred[85]	= 4.371;
        ePhred[86]	= 4.190;
        ePhred[87]	= 4.401;
        ePhred[88]	= 4.171;
        ePhred[89]	= 4.370;
        ePhred[90]	= 4.212;
        ePhred[91]	= 4.415;
        ePhred[92]	= 4.190;
        ePhred[93]	= 4.458;
        ePhred[94]	= 4.135;
        ePhred[95]	= 4.513;
        ePhred[96]	= 4.196;
        ePhred[97]	= 4.492;
        ePhred[98]	= 4.225;
        ePhred[99]	= 4.494;
        ePhred[100]	= 4.224;
        ePhred[101]	= 4.489;
        ePhred[102]	= 4.272;
        ePhred[103]	= 4.533;
        ePhred[104]	= 4.177;
        ePhred[105]	= 4.582;
        ePhred[106]	= 4.230;
        ePhred[107]	= 4.442;
        ePhred[108]	= 4.263;
        ePhred[109]	= 4.337;
        ePhred[110]	= 4.272;
        ePhred[111]	= 4.340;
        ePhred[112]	= 4.315;
        ePhred[113]	= 4.245;
        ePhred[114]	= 4.246;
        ePhred[115]	= 4.195;
        ePhred[116]	= 4.292;
        ePhred[117]	= 4.195;
        ePhred[118]	= 4.344;
        ePhred[119]	= 4.212;
        ePhred[120]	= 4.333;
        ePhred[121]	= 4.218;
        ePhred[122]	= 4.377;
        ePhred[123]	= 4.186;
        ePhred[124]	= 4.333;
        ePhred[125]	= 4.229;
        ePhred[126]	= 4.376;
        ePhred[127]	= 4.234;
        ePhred[128]	= 4.421;
        ePhred[129]	= 4.254;
        ePhred[130]	= 4.422;
        ePhred[131]	= 4.222;
        ePhred[132]	= 4.415;
        ePhred[133]	= 4.241;
        ePhred[134]	= 4.430;
        ePhred[135]	= 4.285;
        ePhred[136]	= 4.473;
        ePhred[137]	= 4.295;
        ePhred[138]	= 4.414;
        ePhred[139]	= 4.312;
        ePhred[140]	= 4.333;
        ePhred[141]	= 4.296;
        ePhred[142]	= 4.334;
        ePhred[143]	= 4.317;
        ePhred[144]	= 4.166;
        ePhred[145]	= 4.356;
        ePhred[146]	= 4.206;
        ePhred[147]	= 4.372;
        ePhred[148]	= 4.247;
        ePhred[149]	= 4.356;
        ePhred[150]	= 4.259;
        ePhred[151]	= 4.383;
        ePhred[152]	= 4.204;
        ePhred[153]	= 4.406;
        ePhred[154]	= 4.224;
        ePhred[155]	= 4.443;
        ePhred[156]	= 4.262;
        ePhred[157]	= 4.424;
        ePhred[158]	= 4.249;
        ePhred[159]	= 4.446;
        ePhred[160]	= 4.265;
        ePhred[161]	= 4.480;
        ePhred[162]	= 4.275;
        ePhred[163]	= 4.504;
        ePhred[164]	= 4.297;
        ePhred[165]	= 4.539;
        ePhred[166]	= 4.291;
        ePhred[167]	= 4.449;
        ePhred[168]	= 4.323;
        ePhred[169]	= 4.402;
        ePhred[170]	= 4.323;
        ePhred[171]	= 4.343;
        ePhred[172]	= 4.357;
        ePhred[173]	= 4.310;
        ePhred[174]	= 4.346;
        ePhred[175]	= 4.321;
        ePhred[176]	= 4.268;
        ePhred[177]	= 4.293;
        ePhred[178]	= 4.300;
        ePhred[179]	= 4.315;
        ePhred[180]	= 4.305;
        ePhred[181]	= 4.347;
        ePhred[182]	= 4.310;
        ePhred[183]	= 4.372;
        ePhred[184]	= 4.334;
        ePhred[185]	= 4.345;
        ePhred[186]	= 4.366;
        ePhred[187]	= 4.366;
        ePhred[188]	= 4.393;
        ePhred[189]	= 4.291;
        ePhred[190]	= 4.384;
        ePhred[191]	= 4.284;
        ePhred[192]	= 4.415;
        ePhred[193]	= 4.299;
        ePhred[194]	= 4.320;
        ePhred[195]	= 4.330;
        ePhred[196]	= 4.277;
        ePhred[197]	= 4.352;
        ePhred[198]	= 4.257;
        ePhred[199]	= 4.348;
        ePhred[200]	= 4.254;
        ePhred[201]	= 4.373;
        ePhred[202]	= 4.246;
        ePhred[203]	= 4.390;
        ePhred[204]	= 4.270;
        ePhred[205]	= 4.413;
        ePhred[206]	= 4.300;
        ePhred[207]	= 4.417;
        ePhred[208]	= 4.187;
        ePhred[209]	= 4.442;
        ePhred[210]	= 4.213;
        ePhred[211]	= 4.467;
        ePhred[212]	= 4.237;
        ePhred[213]	= 4.378;
        ePhred[214]	= 4.265;
        ePhred[215]	= 4.383;
        ePhred[216]	= 4.256;
        ePhred[217]	= 4.409;
        ePhred[218]	= 4.273;
        ePhred[219]	= 4.340;
        ePhred[220]	= 4.298;
        ePhred[221]	= 4.296;
        ePhred[222]	= 4.301;
        ePhred[223]	= 4.273;
        ePhred[224]	= 4.328;
        ePhred[225]	= 4.295;
        ePhred[226]	= 4.344;
        ePhred[227]	= 4.220;
        ePhred[228]	= 4.362;
        ePhred[229]	= 4.246;
        ePhred[230]	= 4.371;
        ePhred[231]	= 4.252;
        ePhred[232]	= 4.394;
        ePhred[233]	= 4.250;
        ePhred[234]	= 4.413;
        ePhred[235]	= 4.275;
        ePhred[236]	= 4.432;
        ePhred[237]	= 4.298;
        ePhred[238]	= 4.442;
        ePhred[239]	= 4.287;
        ePhred[240]	= 4.381;
        ePhred[241]	= 4.312;
        ePhred[242]	= 4.362;
        ePhred[243]	= 4.334;
        ePhred[244]	= 4.298;
        ePhred[245]	= 4.357;
        ePhred[246]	= 4.278;
        ePhred[247]	= 4.289;
        ePhred[248]	= 4.244;
        ePhred[249]	= 4.300;
        ePhred[250]	= 4.266;
        ePhred[251]	= 4.324;
        ePhred[252]	= 4.262;
        ePhred[253]	= 4.333;
        ePhred[254]	= 4.274;
        ePhred[255]	= 4.263;
        ePhred[256]	= 4.277;
        ePhred[257]	= 4.277;
        ePhred[258]	= 4.299;
        ePhred[259]	= 4.299;
        ePhred[260]	= 4.321;
        ePhred[261]	= 4.311;
        ePhred[262]	= 4.268;
        ePhred[263]	= 4.334;
        ePhred[264]	= 4.273;
        ePhred[265]	= 4.279;
        ePhred[266]	= 4.294;
        ePhred[267]	= 4.271;
        ePhred[268]	= 4.317;
        ePhred[269]	= 4.238;
        ePhred[270]	= 4.316;
        ePhred[271]	= 4.236;
        ePhred[272]	= 4.338;
        ePhred[273]	= 4.259;
        ePhred[274]	= 4.358;
        ePhred[275]	= 4.257;
        ePhred[276]	= 4.369;
        ePhred[277]	= 4.269;
        ePhred[278]	= 4.385;
        ePhred[279]	= 4.237;
        ePhred[280]	= 4.404;
        ePhred[281]	= 4.242;
        ePhred[282]	= 4.424;
        ePhred[283]	= 4.260;
        ePhred[284]	= 4.437;
        ePhred[285]	= 4.274;
        ePhred[286]	= 4.458;
        ePhred[287]	= 4.282;
        ePhred[288]	= 4.414;
        ePhred[289]	= 4.302;
        ePhred[290]	= 4.305;
        ePhred[291]	= 4.320;
        ePhred[292]	= 4.317;
        ePhred[293]	= 4.326;
        ePhred[294]	= 4.319;
        ePhred[295]	= 4.345;
        ePhred[296]	= 4.336;
        ePhred[297]	= 4.227;
        ePhred[298]	= 4.295;
        ePhred[299]	= 4.238;
        ePhred[300]	= 4.310;
        ePhred[301]	= 4.253;
        ePhred[302]	= 4.328;
        ePhred[303]	= 4.272;
        ePhred[304]	= 4.333;
        ePhred[305]	= 4.292;
        ePhred[306]	= 4.346;
        ePhred[307]	= 4.306;
        ePhred[308]	= 4.366;
        ePhred[309]	= 4.321;
        ePhred[310]	= 4.324;
        ePhred[311]	= 4.263;
        ePhred[312]	= 4.343;
        ePhred[313]	= 4.258;
        ePhred[314]	= 4.356;
        ePhred[315]	= 4.277;
        ePhred[316]	= 4.366;
        ePhred[317]	= 4.243;
        ePhred[318]	= 4.385;
        ePhred[319]	= 4.261;
        ePhred[320]	= 4.403;
        ePhred[321]	= 4.262;
        ePhred[322]	= 4.410;
        ePhred[323]	= 4.281;
        ePhred[324]	= 4.428;
        ePhred[325]	= 4.299;
        ePhred[326]	= 4.447;
        ePhred[327]	= 4.263;
        ePhred[328]	= 4.460;
        ePhred[329]	= 4.278;
        ePhred[330]	= 4.476;
        ePhred[331]	= 4.287;
        ePhred[332]	= 4.430;
        ePhred[333]	= 4.305;
        ePhred[334]	= 4.447;
        ePhred[335]	= 4.322;
        ePhred[336]	= 4.360;
        ePhred[337]	= 4.265;
        ePhred[338]	= 4.378;
        ePhred[339]	= 4.282;
        ePhred[340]	= 4.381;
        ePhred[341]	= 4.300;
        ePhred[342]	= 4.397;
        ePhred[343]	= 4.317;
        ePhred[344]	= 4.367;
        ePhred[345]	= 4.278;
        ePhred[346]	= 4.384;
        ePhred[347]	= 4.296;
        ePhred[348]	= 4.311;
        ePhred[349]	= 4.313;
        ePhred[350]	= 4.326;
        ePhred[351]	= 4.325;
        ePhred[352]	= 4.342;
        ePhred[353]	= 4.304;
        ePhred[354]	= 4.353;
        ePhred[355]	= 4.284;
        ePhred[356]	= 4.328;
        ePhred[357]	= 4.289;
        ePhred[358]	= 4.342;
        ePhred[359]	= 4.302;
        ePhred[360]	= 4.282;
        ePhred[361]	= 4.279;
        ePhred[362]	= 4.299;
        ePhred[363]	= 4.295;
        ePhred[364]	= 4.315;
        ePhred[365]	= 4.302;
        ePhred[366]	= 4.325;
        ePhred[367]	= 4.318;
        ePhred[368]	= 4.342;
        ePhred[369]	= 4.332;
        ePhred[370]	= 4.358;
        ePhred[371]	= 4.309;
        ePhred[372]	= 4.374;
        ePhred[373]	= 4.323;
        ePhred[374]	= 4.369;
        ePhred[375]	= 4.339;
        ePhred[376]	= 4.341;
        ePhred[377]	= 4.295;
        ePhred[378]	= 4.357;
        ePhred[379]	= 4.311;
        ePhred[380]	= 4.293;
        ePhred[381]	= 4.324;
        ePhred[382]	= 4.306;
        ePhred[383]	= 4.336;
        ePhred[384]	= 4.314;
        ePhred[385]	= 4.351;
        ePhred[386]	= 4.330;
        ePhred[387]	= 4.367;
        ePhred[388]	= 4.249;
        ePhred[389]	= 4.339;
        ePhred[390]	= 4.263;
        ePhred[391]	= 4.354;
        ePhred[392]	= 4.278;
        ePhred[393]	= 4.370;
        ePhred[394]	= 4.287;
        ePhred[395]	= 4.368;
        ePhred[396]	= 4.302;
        ePhred[397]	= 4.357;
        ePhred[398]	= 4.312;
        ePhred[399]	= 4.344;
        ePhred[400]	= 4.295;
        ePhred[401]	= 4.352;
        ePhred[402]	= 4.309;
        ePhred[403]	= 4.363;
        ePhred[404]	= 4.321;
        ePhred[405]	= 4.350;
        ePhred[406]	= 4.335;
        ePhred[407]	= 4.364;
        ePhred[408]	= 4.349;
        ePhred[409]	= 4.372;
        ePhred[410]	= 4.361;
        ePhred[411]	= 4.386;
        ePhred[412]	= 4.375;
        ePhred[413]	= 4.400;
        ePhred[414]	= 4.390;
        ePhred[415]	= 4.384;
        ePhred[416]	= 4.381;
        ePhred[417]	= 4.353;
        ePhred[418]	= 4.369;
        ePhred[419]	= 4.367;
        ePhred[420]	= 4.325;
        ePhred[421]	= 4.378;
        ePhred[422]	= 4.303;
        ePhred[423]	= 4.391;
        ePhred[424]	= 4.291;
        ePhred[425]	= 4.405;
        ePhred[426]	= 4.304;
        ePhred[427]	= 4.351;
        ePhred[428]	= 4.313;
        ePhred[429]	= 4.365;
        ePhred[430]	= 4.304;
        ePhred[431]	= 4.347;
        ePhred[432]	= 4.314;
        ePhred[433]	= 4.359;
        ePhred[434]	= 4.327;
        ePhred[435]	= 4.373;
        ePhred[436]	= 4.338;
        ePhred[437]	= 4.368;
        ePhred[438]	= 4.350;
        ePhred[439]	= 4.360;
        ePhred[440]	= 4.364;
        ePhred[441]	= 4.297;
        ePhred[442]	= 4.374;
        ePhred[443]	= 4.310;
        ePhred[444]	= 4.362;
        ePhred[445]	= 4.318;
        ePhred[446]	= 4.375;
        ePhred[447]	= 4.313;
        ePhred[448]	= 4.386;
        ePhred[449]	= 4.322;
        ePhred[450]	= 4.399;
        ePhred[451]	= 4.336;
        ePhred[452]	= 4.412;
        ePhred[453]	= 4.345;
        ePhred[454]	= 4.424;
        ePhred[455]	= 4.288;
        ePhred[456]	= 4.395;
        ePhred[457]	= 4.244;
        ePhred[458]	= 4.383;
        ePhred[459]	= 4.254;
        ePhred[460]	= 4.377;
        ePhred[461]	= 4.267;
        ePhred[462]	= 4.356;
        ePhred[463]	= 4.278;
        ePhred[464]	= 4.369;
        ePhred[465]	= 4.291;
        ePhred[466]	= 4.363;
        ePhred[467]	= 4.255;
        ePhred[468]	= 4.374;
        ePhred[469]	= 4.267;
        ePhred[470]	= 4.384;
        ePhred[471]	= 4.280;
        ePhred[472]	= 4.379;
        ePhred[473]	= 4.267;
        ePhred[474]	= 4.388;
        ePhred[475]	= 4.279;
        ePhred[476]	= 4.400;
        ePhred[477]	= 4.285;
        ePhred[478]	= 4.413;
        ePhred[479]	= 4.275;
        ePhred[480]	= 4.351;
        ePhred[481]	= 4.282;
        ePhred[482]	= 4.363;
        ePhred[483]	= 4.282;
        ePhred[484]	= 4.355;
        ePhred[485]	= 4.291;
        ePhred[486]	= 4.366;
        ePhred[487]	= 4.301;
        ePhred[488]	= 4.378;
        ePhred[489]	= 4.298;
        ePhred[490]	= 4.390;
        ePhred[491]	= 4.308;
        ePhred[492]	= 4.367;
        ePhred[493]	= 4.320;
        ePhred[494]	= 4.379;
        ePhred[495]	= 4.302;
        ePhred[496]	= 4.391;
        ePhred[497]	= 4.272;
        ePhred[498]	= 4.337;
        ePhred[499]	= 4.268;
        return ePhred;
    }

}
