package org.broadinstitute.sting.gatk.walkers.SMUGwalkers;

import net.sf.samtools.SAMReadGroupRecord;
import org.broadinstitute.sting.commandline.Argument;
import org.broadinstitute.sting.commandline.ArgumentCollection;
import org.broadinstitute.sting.commandline.Output;
import org.broadinstitute.sting.commandline.RodBinding;
import org.broadinstitute.sting.gatk.DownsampleType;
import org.broadinstitute.sting.gatk.arguments.DbsnpArgumentCollection;
import org.broadinstitute.sting.gatk.contexts.AlignmentContext;
import org.broadinstitute.sting.gatk.contexts.AlignmentContextUtils;
import org.broadinstitute.sting.gatk.contexts.ReferenceContext;
import org.broadinstitute.sting.gatk.filters.BadMateFilter;
import org.broadinstitute.sting.gatk.filters.MappingQualityUnavailableFilter;
import org.broadinstitute.sting.gatk.refdata.RefMetaDataTracker;
import org.broadinstitute.sting.gatk.walkers.*;
import org.broadinstitute.sting.gatk.walkers.annotator.VariantAnnotatorEngine;
import org.broadinstitute.sting.gatk.walkers.annotator.interfaces.AnnotatorCompatibleWalker;
import org.broadinstitute.sting.gatk.walkers.genotyper.UnifiedArgumentCollection;
import org.broadinstitute.sting.gatk.walkers.genotyper.UnifiedGenotyperEngine;
import org.broadinstitute.sting.gatk.walkers.genotyper.VariantCallContext;
import org.broadinstitute.sting.utils.BaseUtils;
import org.broadinstitute.sting.utils.GenomeLoc;
import org.broadinstitute.sting.utils.pileup.ReadBackedPileup;
import org.broadinstitute.sting.utils.variantcontext.Genotype;
import org.broadinstitute.sting.utils.variantcontext.VariantContext;
import org.broadinstitute.sting.utils.variantcontext.VariantContextUtils;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.*;

/**
 * Created by Zhuo Song and Chun Li.
 * User: zsong
 * Date: 5/22/12
 * Time: 12:12 PM
 * To change this template use File | Settings | File Templates.
 */
@ReadFilters( {BadMateFilter.class, MappingQualityUnavailableFilter.class} )
@Reference(window=@Window(start=-200,stop=200))
@By(DataSource.REFERENCE)
@Downsample(by= DownsampleType.BY_SAMPLE, toCoverage=250)
public class SMUG_BaseSub extends LocusWalker<Integer, Integer> implements AnnotatorCompatibleWalker {

    @Output
    PrintStream out;

    @ArgumentCollection
    private UnifiedArgumentCollection UAC = new UnifiedArgumentCollection();
    @ArgumentCollection
    protected DbsnpArgumentCollection dbsnp = new DbsnpArgumentCollection();
    public RodBinding<VariantContext> getDbsnpRodBinding() { return dbsnp.dbsnp; }
    public RodBinding<VariantContext> getSnpEffRodBinding() { return null; }
    public List<RodBinding<VariantContext>> getCompRodBindings() { return Collections.emptyList(); }
    public List<RodBinding<VariantContext>> getResourceRodBindings() { return Collections.emptyList(); }
    public boolean alwaysAppendDbsnpId() { return false; }

    @Argument(fullName = "debug_file", shortName = "debug_file", doc = "File to print all of the annotated and detailed debugging output", required = false)
    protected PrintStream verboseWriter = null;
    @Argument(fullName="annotation", shortName="A", doc="One or more specific annotations to apply to variant calls", required=false)
    protected List<String> annotationsToUse = new ArrayList<String>();
    @Argument(fullName="excludeAnnotation", shortName="XA", doc="One or more specific annotations to exclude", required=false)
    protected List<String> annotationsToExclude = new ArrayList<String>();
    @Argument(fullName="group", shortName="G", doc="One or more classes/groups of annotations to apply to variant calls", required=false)
    protected String[] annotationClassesToUse = { "Standard" };
    @Argument(fullName = "SampleName_Tumor", shortName = "smT", doc = "sample name of tumor tissue", required = true)
    String smT;
    @Argument(fullName = "SampleName_Normal", shortName = "smN", doc = "sample name of normal tissue", required = true)
    String smN;
    @Argument(fullName = "SampleType_Tumor", shortName = "typeT", doc = "sample type of tumor tissue", required = false)
    String typeT       = "Tumor";
    @Argument(fullName = "SampleType_Normal", shortName = "typeN", doc = "sample type of normal tissue", required = false)
    String typeN       = "Normal";
    @Argument(fullName = "minBaseQ", shortName = "minBQ", doc = "min base quality, default=20", required = false)
    int minBQ          = 20;
    @Argument(fullName = "minMappingQ",  shortName = "minMQ", doc = "min mapping quality, default=20", required = false)
    int minMQ          = 20;
    @Argument(fullName = "minAllAlleles_Normal", shortName = "minAllN", doc = "min sum of all base counts in Normal after completely filtering (BQ,MQ) of the sample, default=10", required = false)
    int minAllNormal   = 10;
    @Argument(fullName = "minAllAlleles_Tumor", shortName = "minAllT", doc = "min sum of all base counts in Tumor after completely filtering (BQ,MQ) of the sample, default=5", required = false)
    int minAllTumor    = 10;
    @Argument(fullName = "minMinorAlleles_Tumor", shortName = "minRestT", doc = "min sum of three minor alleles' base counts in Tumor after completely filtering (BQ,MQ) of the sample, default=2", required = false)
    int minRestTumor   = 2;
    @Argument(fullName = "minHeteroRate_Tumor", shortName = "minHRT", doc = "min hetero rate in Tumor: Minors/All, default=0.1", required = false)
    double minHRT      = 0.01;
    /*   GATK:
         Since we cap each base quality (BQ) by its read's mapping quality (MQ) [in Read.updateBaseAndQuality()], then:
         if minBQ > minMQ, then we require that MQ be >= minBQ as well.
         [Otherwise, we end up capping BQ by MQ only AFTER we tried removing bases with BQ < minBQ, which is WRONG!]
         To do this properly, we set: minMQ = max(minMQ, minBQ)
    */
    protected PrintStream BaseSub = null;
    private Set<String> _sampleNames = new HashSet<String>();
    private Set<String> _sampleName1 = new HashSet<String>();
    private Set<String> _sampleName2 = new HashSet<String>();
    private int _isTExist = 0;
    private int _isBExist = 0;
    private String _BSf = null;
    /*  GATK:
        //the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls. This argument is passed to the Unified Genotyper.
        @Argument(fullName="standard_min_confidence_threshold_for_calling", shortName="stand_call_conf", doc="the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls", required=false)
        private double callConf = -1;
        //@Argument(fullName ="standard_min_confidence_threshold_for_calling", shortName="callConf", doc="the minimum phred-scaled Qscore threshold to separate high confidence from low confidence calls, default=30", required=false)
        //double callConf    = 30;
    */
    private UnifiedGenotyperEngine UG_engine1 = null;
    private UnifiedGenotyperEngine UG_engine2 = null;
    private VariantAnnotatorEngine annotationEngine;
    double call_Conf = 0.0;
    int    call_mbq  = 0;

    public void initialize() {
        call_Conf = UAC.STANDARD_CONFIDENCE_FOR_CALLING;
        call_mbq  = UAC.MIN_BASE_QUALTY_SCORE;
        List<SAMReadGroupRecord> read_groups = this.getToolkit().getSAMFileHeader().getReadGroups();
        for ( SAMReadGroupRecord record : read_groups ) {
            String sample = record.getSample();
            if ( sample != null )
                _sampleNames.add(sample);
        }
        for ( String sample : _sampleNames) {
            if (sample.equals(smT)) {
                _isTExist = 1;
                _sampleName1.add(smT);
            }
            if (sample.equals(smN)) {
                _isBExist = 1;
                _sampleName2.add(smN);
            }
            //out.println("## " + sample);
        }
        String _error1 = "## ERROR 1:\t" + smT + " doesn't exist in data";
        if (_isTExist == 0) {
            System.err.println(_error1);
            System.exit(0);
        }
        String _error2 = "## ERROR 2:\t" + smN + " doesn't exist in data";
        if (_isBExist == 0) {
            System.err.println(_error2);
            System.exit(0);
        }
        //
        if ( verboseWriter != null )
            verboseWriter.println("AFINFO\tLOC\tREF\tALT\tMAF\tF\tAFprior\tAFposterior\tNormalizedPosterior");
        annotationEngine = new VariantAnnotatorEngine(Arrays.asList(annotationClassesToUse), annotationsToUse, annotationsToExclude, this, getToolkit());
        UG_engine1 = new UnifiedGenotyperEngine(getToolkit(), UAC, logger, verboseWriter, annotationEngine, _sampleName1, VariantContextUtils.DEFAULT_PLOIDY);
        UG_engine2 = new UnifiedGenotyperEngine(getToolkit(), UAC, logger, verboseWriter, annotationEngine, _sampleName2, VariantContextUtils.DEFAULT_PLOIDY);
        //
        _BSf  = smN + "." + typeN + "." + smT + "." + typeT + "." + "ToAnalyzeBaseSub.txt";
        //
        FileOutputStream _BS = null;
        try {
            _BS = new FileOutputStream(_BSf);
        } catch (FileNotFoundException e) {
            e.printStackTrace(System.err);
        }
        BaseSub = new PrintStream(_BS);
        String BaseSubHeader = "#loc\trsID\tref\tgt_N\tgt_T\tgtQ_N\tgtQ_T\tN_hR\tT_hR\tN_hR_C\tT_hR_C\tN_aC\tT_aC\tN_+-\tT_+-\tN_filtered\tT_filtered";
        BaseSub.println(BaseSubHeader);

    }

    @Override
    public Integer map(RefMetaDataTracker tracker, ReferenceContext ref, AlignmentContext context) {
        int _nPotentialBaseSub = 0;
        DecimalFormat df4 = new DecimalFormat("#.####");
        DecimalFormat df2 = new DecimalFormat("#.##");
        //
        Map<String, AlignmentContext> _smT_ContextMap;
        Map<String, AlignmentContext> _smN_ContextMap;
        //
        if ( BaseUtils.simpleBaseToBaseIndex(ref.getBase()) != -1 ) {
            if ( ! context.hasExtendedEventPileup() ) {
                _smT_ContextMap = AlignmentContextUtils.splitContextBySampleName(context, smT);
                _smN_ContextMap = AlignmentContextUtils.splitContextBySampleName(context, smN);
                if (_smT_ContextMap != null && _smN_ContextMap != null) {
                    // before filter BQ, MQ; GenotypeCall based on this
                    AlignmentContext _smT_Context = _smT_ContextMap.get(smT);
                    AlignmentContext _smN_Context = _smN_ContextMap.get(smN);
                    //
                    if (_smT_Context != null && _smN_Context != null) {
                        ReadBackedPileup _smT_Context_COMPLETE_RBP = _smT_Context.getBasePileup();
                        ReadBackedPileup _smN_Context_COMPLETE_RBP = _smN_Context.getBasePileup();
                        ReadBackedPileup _smT_Context_FORWARD_RBP  = _smT_Context.getBasePileup().getPositiveStrandPileup();
                        ReadBackedPileup _smN_Context_FORWARD_RBP  = _smN_Context.getBasePileup().getPositiveStrandPileup();
                        ReadBackedPileup _smT_Context_REVERSE_RBP  = _smT_Context.getBasePileup().getNegativeStrandPileup();
                        ReadBackedPileup _smN_Context_REVERSE_RBP  = _smN_Context.getBasePileup().getNegativeStrandPileup();
                        int _nsmTcc = _smT_Context_COMPLETE_RBP.depthOfCoverage();
                        int _nsmNcc = _smN_Context_COMPLETE_RBP.depthOfCoverage();
                        GenomeLoc _location = _smT_Context_COMPLETE_RBP.getLocation();
                        //
                        if (_nsmTcc >= 1 && _nsmNcc >= 1) {
                            ReadBackedPileup _smT_COMPLETE_Filtered_BQMQ = _smT_Context_COMPLETE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _smN_COMPLETE_Filtered_BQMQ = _smN_Context_COMPLETE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_CompleteF                          = _smT_COMPLETE_Filtered_BQMQ.depthOfCoverage();
                            int _nsmN_CompleteF                          = _smN_COMPLETE_Filtered_BQMQ.depthOfCoverage();
                            int[] _smT_CompleteFBC                       = _smT_COMPLETE_Filtered_BQMQ.getBaseCounts();
                            int[] _smN_CompleteFBC                       = _smN_COMPLETE_Filtered_BQMQ.getBaseCounts();
                            int _aT = _smT_CompleteFBC[0], _cT = _smT_CompleteFBC[1], _gT = _smT_CompleteFBC[2], _tT = _smT_CompleteFBC[3];
                            int _aN = _smN_CompleteFBC[0], _cN = _smN_CompleteFBC[1], _gN = _smN_CompleteFBC[2], _tN = _smN_CompleteFBC[3];
                            ReadBackedPileup _SM1_FORWARD_FilteredBQMQ = _smT_Context_FORWARD_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _SM2_FORWARD_FilteredBQMQ = _smN_Context_FORWARD_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_ForwardF                         = _SM1_FORWARD_FilteredBQMQ.depthOfCoverage();
                            int _nsmN_ForwardF                         = _SM2_FORWARD_FilteredBQMQ.depthOfCoverage();
                            int[] _nsmT_ForwardFBC                     = _SM1_FORWARD_FilteredBQMQ.getBaseCounts();
                            int[] _nsmN_ForwardFBC                     = _SM2_FORWARD_FilteredBQMQ.getBaseCounts();
                            int _afT = _nsmT_ForwardFBC[0], _cfT = _nsmT_ForwardFBC[1], _gfT = _nsmT_ForwardFBC[2], _tfT = _nsmT_ForwardFBC[3];
                            int _afN = _nsmN_ForwardFBC[0], _cfN = _nsmN_ForwardFBC[1], _gfN = _nsmN_ForwardFBC[2], _tfN = _nsmN_ForwardFBC[3];
                            ReadBackedPileup _smT_REVERSE_FilteredBQMQ = _smT_Context_REVERSE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            ReadBackedPileup _smN_REVERSE_FilteredBQMQ = _smN_Context_REVERSE_RBP.getBaseAndMappingFilteredPileup(minBQ, minMQ);
                            int _nsmT_ReverseF                         = _smT_REVERSE_FilteredBQMQ.depthOfCoverage();
                            int _nsmN_ReverseF                         = _smN_REVERSE_FilteredBQMQ.depthOfCoverage();
                            int[] _nsmT_ReverseFBC                     = _smT_REVERSE_FilteredBQMQ.getBaseCounts();
                            int[] _nsmN_ReverseFBC                     = _smN_REVERSE_FilteredBQMQ.getBaseCounts();
                            int _arT = _nsmT_ReverseFBC[0], _crT = _nsmT_ReverseFBC[1], _grT = _nsmT_ReverseFBC[2], _trT = _nsmT_ReverseFBC[3];
                            int _arN = _nsmN_ReverseFBC[0], _crN = _nsmN_ReverseFBC[1], _grN = _nsmN_ReverseFBC[2], _trN = _nsmN_ReverseFBC[3];
                            int[] _T  = {_aT,_cT,_gT,_tT};
                            int[] _N  = {_aN,_cN,_gN,_tN};
                            int _All_T = 0;
                            int _All_N = 0;
                            int i;
                            for (i = 0; i < 4; i++) {
                                _All_T += _T[i];
                                _All_N += _N[i];
                            }
                            //
                            AlignmentContext _smT_Context_BQMQ = new AlignmentContext(_location, _smT_COMPLETE_Filtered_BQMQ);
                            AlignmentContext _smN_Context_BQMQ = new AlignmentContext(_location, _smN_COMPLETE_Filtered_BQMQ);
                            //VariantCallContext _vcT = UG_engine1.calculateLikelihoodsAndGenotypes(tracker, ref, _smT_Context);
                            //VariantCallContext _vcN = UG_engine2.calculateLikelihoodsAndGenotypes(tracker, ref, _smN_Context);
                            VariantCallContext _vcT = UG_engine1.calculateLikelihoodsAndGenotypes(tracker, ref, _smT_Context_BQMQ).get(0);
                            VariantCallContext _vcN = UG_engine2.calculateLikelihoodsAndGenotypes(tracker, ref, _smN_Context_BQMQ).get(0);
                            //
                            String _refAllele = new String(new byte[]{ref.getBase()});
                            String _rsID = ".";
                            if (_vcT.hasID()) {
                                _rsID = _vcT.getID();
                            }
                            /********************/
                            //String _s0 = _location+"\t"+_rsID+"\t"+_refAllele+"\t"+_aN+","+_cN+","+_gN+","+_tN+"\t"+_aT+","+_cT+","+_gT+","+_tT+"\t"+_vcN.toString()+"\t"+_vcT.toString();
                            //out.println(_s0);
                            /********************/
                            Genotype _gt_N = null;
                            String  _gtS_N = "%";
                            double  _gtQ_N = -100.0;
                            if (_vcN.hasGenotype(smN)) {
                                 _gt_N =  _vcN.getGenotype(smN);
                                _gtS_N = _gt_N.getGenotypeString(false);
                                _gtQ_N = _gt_N.getPhredScaledQual();
                            }
                            Genotype _gt_T = null;
                            String  _gtS_T = "%";
                            double  _gtQ_T = -100.0;
                            if (_vcT.hasGenotype(smT)) {
                                 _gt_T =  _vcT.getGenotype(smT);
                                _gtS_T = _gt_T.getGenotypeString(false);
                                _gtQ_T = _gt_T.getPhredScaledQual();
                            }
                            //
                            String _lineGT = _location
                                    + "\t" + _rsID
                                    + "\t" + _refAllele
                                    + "\t" + _gtS_N
                                    + "\t" + _gtS_T
                                    + "\t" + df2.format(_gtQ_N)
                                    + "\t" + df2.format(_gtQ_T)
                                    ;
                            /********** BaseSub **********/
                            Map<String, Integer> map_T = new HashMap<String, Integer>();
                            Map<String, Integer> map_N = new HashMap<String, Integer>();
                            map_T.put("A",_aT);
                            map_T.put("C",_cT);
                            map_T.put("G",_gT);
                            map_T.put("T",_tT);
                            map_N.put("A",_aN);
                            map_N.put("C",_cN);
                            map_N.put("G",_gN);
                            map_N.put("T",_tN);
                            int BaseSub_Report   = 0;
                            boolean isNormalHomo = false;
                            boolean isTumorVar   = false;
                            String _allele_N     = "*";
                            String N_homo        = "N";
                            if (_All_T >= minAllTumor && _All_N >= minAllNormal) {
                                if (_vcN.getNAlleles() == 1) {
                                    isNormalHomo = true;
                                    // _allele_N = C*
                                    _allele_N = _vcN.getAlleles().get(0).toString();
                                } else if (_vcN.getNAlleles() > 1) {
                                    if (_vcN.hasGenotype(smN)) {
                                        if (_gt_N.isHom() && _gtQ_N >= call_Conf) {
                                            isNormalHomo = true;
                                            // _allele_N = G/G
                                            _allele_N = _gtS_N;
                                        }
                                    } else {
                                        out.println("## WEIRD: "+_location+","+_rsID+" (Normal) with >1 allels but doesn't have genotype ##");
                                    }
                                }
                            }
                            // genotype of Normal could be '%' NO_VARIATION or 'G/G' high qual homo, but not '.' NO_CALL
                            if (isNormalHomo) {
                                N_homo = _allele_N.substring(0,1);
                                //out.println(_allele_N+"("+N_homo+")");
                                int _varChrCount_N = 0;
                                int _allChrCount_N = 0;
                                for (String key : map_N.keySet()) {
                                    if (!key.equalsIgnoreCase(N_homo)) {
                                        _varChrCount_N += map_N.get(key);
                                    }
                                    _allChrCount_N += map_N.get(key);
                                }
                                int _varChrCount_T = 0;
                                int _allChrCount_T = 0;
                                for (String key : map_T.keySet()) {
                                    if (!key.equalsIgnoreCase(N_homo)) {
                                        _varChrCount_T += map_T.get(key);
                                    }
                                    _allChrCount_T += map_T.get(key);
                                }
                                //out.println(_varChrCount_N+"/"+_allChrCount_N+"\t"+_varChrCount_T+"/"+_allChrCount_T);
                                //
                                double _HeteroR_T = 0.0;
                                double _HeteroR_N = 0.0;
                                _HeteroR_T = (double)_varChrCount_T/(double)_allChrCount_T;
                                _HeteroR_N = (double)_varChrCount_N/(double)_allChrCount_N;
                                if (_varChrCount_T >= minRestTumor && _HeteroR_T >= minHRT) {
                                    isTumorVar = true;
                                }
                                if (isTumorVar) {
                                    BaseSub_Report = 1;
                                }
                                //
                                if (BaseSub_Report > 0) {
                                    String _lineBaseSub = df4.format(_HeteroR_N)
                                            + "\t" + df4.format(_HeteroR_T)
                                            + "\t" + _varChrCount_N + "/" + _allChrCount_N
                                            + "\t" + _varChrCount_T + "/" + _allChrCount_T
                                            + "\t" + _aN + "," + _cN + "," + _gN + "," + _tN
                                            + "\t" + _aT + "," + _cT + "," + _gT + "," + _tT
                                            + "\t" + _afN + "," + _arN + ";" + _cfN + "," + _crN + ";" + _gfN + "," + _grN + ";" + _tfN + "," + _trN
                                            + "\t" + _afT + "," + _arT + ";" + _cfT + "," + _crT + ";" + _gfT + "," + _grT + ";" + _tfT + "," + _trT
                                            + "\t" + smN + "(" + typeN + "):" + _nsmNcc + "->" + _nsmN_CompleteF + "=" + "(" + _nsmN_ForwardF + "+," + _nsmN_ReverseF + "-)"
                                            + "\t" + smT + "(" + typeT + "):" + _nsmTcc + "->" + _nsmT_CompleteF + "=" + "(" + _nsmT_ForwardF + "+," + _nsmT_ReverseF + "-)"
                                            ;
                                    String _outLine = _lineGT + "\t" + _lineBaseSub;
                                    BaseSub.println(_outLine);
                                    _nPotentialBaseSub++;
                                }
                            }

                        }
                    }
                }
            }
        }
        return _nPotentialBaseSub;
    }

    @Override
    public Integer reduceInit() {
        return 0;  //To change body of implemented methods use File | Settings | File Templates.
    }

    @Override
    public Integer reduce(Integer value, Integer sum) {
        return value + sum;  //To change body of implemented methods use File | Settings | File Templates.
    }

    public void onTraversalDone(Integer result) {
        out.println("## SMUG ## ---------------------------------------------------------------------");
        out.printf( "## SMUG ## typeN %s, typeT %s, minBQ %d, minMQ %d, minAllNormal %d, minAllTumor %d, minRestTumor %d, minHRT %.2f, callConf %.2f%n", typeN, typeT, minBQ, minMQ, minAllNormal, minAllTumor, minRestTumor, minHRT, call_Conf);
        out.println("## SMUG ## " + _BSf);
        out.println("## SMUG ## The No. of loci with potential NormalHomo TumorHetero (BaseSub) is: " + result);
        out.println("## SMUG ## ---------------------------------------------------------------------");
    }
}
